% Hybrid A* path search based on code for the A* algorithm by: 
% Premakumar, P.(2016): A* (A Star) search for path planning tutorial. 
% Version 1.2.0.1. Available at: https://se.mathworks.com/matlabcentral/
% fileexchange/26248-a-a-star-search-for-path-planning-tutorial (accessed:
% 01.11.2019)

% Cost function based on information provided by the BINN guidance model
% Penalty on heading changes equal to the heading in radians
function [psi_p_new, pos_new, idx_new, optimal_path] = hybridAstarBINN(Z, p0,...
    psi0, Nx, Ny, dX, dY, Lx, Ly, x, y, plot_search, plot_path, heuristics)

    %% Hybrid A* parameters
    d = norm([dX,dY])+1e-5; %[m] Driving distance (suffices to leave cell) 
    dPsi = 10* pi/180; %[rad] Resolution of the heading angle (10deg)
    turnAngles = [-3*dPsi,0,3*dPsi]'; %[rad] Turn angles defining 3 successors
    posAccuracy = max(dX,dY); %[m] Radius of acceptance
    %headingAccuracy = dPsi; %[rad] Accepted deviation in heading %%NB! No specified desired heading at the target%%

    %% Obstacles and Targets
    % Target
    [~,idx_t] = min(Z(:)); % Find target based on neural activity
    targetCell = zeros(1,2); % Cell containing the target
    [targetCell(1),targetCell(2)] = findCell(x(idx_t),y(idx_t),dX,dY);
    
    % Obstacles
    idx_o = find(Z>0);  % Find obstacles based on neural activity
    No = length(idx_o); % Number of obstacles
    obstacleCell = zeros(No,2);
    obstacleCell(:,1) = floor(idx_o/Ny)+1; % Obstructed cells
    obstacleCell(:,2) = idx_o-Ny*(obstacleCell(:,1)-1);

    %% Lists used for the A* algorithm
    % Open List Structure
    % --------------------------------------------------------------------
    % |is on list 1/0|xCell|yCell|psi|xPos|yPos|idx_parent|h(n)|g(n)|f(n)|
    % --------------------------------------------------------------------
    OPEN = zeros(10000,10); % Preallocate memory
    % Closed List Structure
    % -----------------------------------
    % |xCell|yCell|noClosedPsi|closedPsi|
    % -----------------------------------
    CLOSED = zeros(10000,3+floor(2*pi/dPsi)+1); % Preallocate memory
    
    CLOSED(1:No,:) = [obstacleCell,... % Insert all obstructed cells (close for all headings)
        repmat([floor(2*pi/dPsi)+1,(-pi:dPsi:pi)],No,1)];
    CLOSED_count = No; % # nodes in closed list

    % Initial state
    xPos = p0(1); yPos = p0(2); % Initial node
    [xCell,yCell] = findCell(xPos, yPos, dX, dY); % Intial cell
    psi = roundValue(psi0,dPsi); % [rad] Initial heading (discrete)
    path_cost = 0; % Cost along expanded path
    if heuristics % Cost to target
        % Euclidean distance adjusted for neural activity
        gn = 0.5*(Z(targetCell(2),targetCell(1))+1)*... 
        distance(xPos, yPos, x(idx_t), y(idx_t));
    else
        % Euclidean distance
        gn = distance(xPos, yPos, x(idx_t), y(idx_t)); 
    end
    OPEN_count = 1; % # nodes in open list
    OPEN(OPEN_count,:) = [1,xCell,yCell,psi,xPos,yPos,...
        OPEN_count,path_cost,gn,gn]; % Insert to open list
    OPEN(OPEN_count,1) = 0; % Remove from open list    
    CLOSED_count = CLOSED_count + 1; 
    CLOSED(CLOSED_count,1:4) = [xCell,yCell,1,psi]; % Insert to closed list
    
    if plot_search % Initialization of search tree plot
        figure(21); clf; hold on;      
        %plot(y,x,'.','Color',0.8*[1 1 1],'Linewidth',2); % Nodes
        plot(y(idx_t),x(idx_t),'p','Color',[0.6350,0.0780,0.1840],...% Target  
            'MarkerFaceColor',[0.6350,0.0780,0.1840],'Markersize',8); 
        plot(y(idx_o),x(idx_o),'ks','MarkerFaceColor','k'); %Obstacles
        [yGrid,xGrid] = meshgrid(0:dY:Ly,0:dX:Lx); % Grid
        plot(yGrid,xGrid,':','Color',0.8*[1 1 1]);
        plot(yGrid',xGrid',':','Color',0.8*[1 1 1]);
        xlabel('y [m]'); ylabel('x [m]');
        axis equal tight, box on
        hold off; 
    end
    
    %% Hybrid A* algorithm
    idx_OPEN_min = OPEN_count;
    noPath = 0; % there exists no path = false
    while( norm([x(idx_t)-xPos,y(idx_t)-yPos]) > posAccuracy && ~noPath )

        % Expanded Array Structure
        % ------------------------------------------
        % |xCell|yCell|psi|xPos|yPos|h(n)|g(n)|f(n)|
        % ------------------------------------------ 
        exp_array = expand_array(psi, xPos, yPos, path_cost, x(idx_t),...
            y(idx_t), CLOSED(1:CLOSED_count,:), Lx, Ly, dX, dY,...
            turnAngles, d, dPsi, Z, heuristics);
        exp_num = size(exp_array,1);
        cut = 1:exp_num;
        for i = 1:exp_num
            flag = 0;
            for j = 1:OPEN_count
                if(abs(exp_array(i,4) - OPEN(j,5)) < 1e-5 ...
                        && abs(exp_array(i,5) - OPEN(j,6)) < 1e-5)
                    % Node is already expanded from another direction
                    flag = 1;
                elseif(exp_array(i,1) == OPEN(j,2) && exp_array(i,2)...
                      == OPEN(j,3) && abs(exp_array(i,3)-OPEN(j,4)) < 1e-5)
                    % Cell is already approached with this angle (keep the cheapest)
                    OPEN(j,10) = min(OPEN(j,10),exp_array(i,8));
                    if OPEN(j,10) == exp_array(i,8)
                        % Update parents, gn, hn
                        OPEN(j,5:10) = [exp_array(i,4:5),idx_OPEN_min,...
                            exp_array(i,6:8)];
                        cut = setdiff(cut,i);
                    end % End of check for minimum fn
                    flag = 1;
                end 
            end % End of checking if child node already exists in open list 
            if ~flag
                OPEN_count = OPEN_count + 1;
                OPEN(OPEN_count,:) = [1,exp_array(i,1:5),idx_OPEN_min,...
                    exp_array(i,6:8)];
                cut = setdiff(cut,i); % Draw this expansion in search tree
            end % End of inserting new element into the open list        
        end % End of i loop
        exp_array(cut,:) = []; % Cut non-expanded nodes
        [exp_num,~] = size(exp_array);

        if plot_search % Plot search tree
            figure(21); hold on;
            plot(yPos,xPos,'.','MarkerSize',10,'Color',1.2*[0.4660 0.6740 0.1880])%'m', 6 is default MarkerSize
            plot(exp_array(:,5),exp_array(:,4),'.','Color',1.2*[0.4660 0.6740 0.1880])%'m'
            plot([repmat(yPos,exp_num,1),exp_array(:,5)]', ...
                [repmat(xPos,exp_num,1),exp_array(:,4)]','Color',1.2*[0.4660 0.6740 0.1880])%rand(1,3))
            hold off;
        end

        % Find node with the smallest total cost, fn
        idx_OPEN_min = min_fn(OPEN,OPEN_count,x(idx_t),y(idx_t),posAccuracy);
        if(idx_OPEN_min ~= -1) % If there exists a feasible path:
            % Select node with the smallest total cost, fn, or the target
            xCell = OPEN(idx_OPEN_min,2);
            yCell = OPEN(idx_OPEN_min,3);
            psi = OPEN(idx_OPEN_min,4);
            xPos = OPEN(idx_OPEN_min,5);
            yPos = OPEN(idx_OPEN_min,6);
            % Update the cost of reaching the parent node
            path_cost = OPEN(idx_OPEN_min,8);
            % Move the node to the CLOSED list
            [inCLOSED,idx_CLOSED] = ismember([xCell yCell],CLOSED(:,1:2),'rows');
            if inCLOSED
                CLOSED(idx_CLOSED,3) = CLOSED(idx_CLOSED,3)+1;
                CLOSED(idx_CLOSED,CLOSED(idx_CLOSED,3)+3) = psi;
            else
                CLOSED_count = CLOSED_count + 1;
                CLOSED(CLOSED_count,1:4) = [xCell yCell 1 psi];
            end
            OPEN(idx_OPEN_min,1) = 0;
        else % There does not exist a path to the target
            noPath = 1; % Exits the loop!
        end 

    end
    
    if(noPath == true) % No feasible path: Change operation mode to DP
        fprintf('No feasible path to the target: Operation mode is changed to DP.\n');
        pos_new = p0';
        [xCell,yCell] = findCell(p0(1),p0(2),dX,dY);
        idx_new = (xCell-1)*Ny + yCell;
        psi_p_new = psi0;
        optimal_path = p0;
        return;
    else % Traverse the open list and extract the path 
        optimal_path = zeros(Nx*Ny,2);
        i = 1;
        % Adds target to the path instead of the last expanded node
        optimal_path(i,:) = [x(idx_t) y(idx_t)]; 
        parent_idx = OPEN(idx_OPEN_min,7);
        while(OPEN(parent_idx,5) ~= p0(1) || OPEN(parent_idx,6) ~= p0(2))
            i = i + 1;
            optimal_path(i,:) = OPEN(parent_idx,5:6);
            parent_idx = OPEN(parent_idx,7); % Grandparent node
        end
        i = i + 1;
        optimal_path(i,:) = p0;
        optimal_path = optimal_path(1:i,:);
    end 
    pos_new = optimal_path(i-1,1:2);
    [xCell,yCell] = findCell(pos_new(1),pos_new(2),dX,dY);
    idx_new = (xCell-1)*Ny + yCell;
    psi_p_new = atan2(pos_new(2)-p0(2),pos_new(1)-p0(1));
    
    if plot_search % Plot final path
        figure(21); hold on;
        plot(optimal_path(:,2),optimal_path(:,1),'.-',... % Optimal path
            'Color',[0.4660 0.6740 0.1880],'LineWidth',1.5,'MarkerSize',12); 
        hold off;
    end

    if plot_path % Plot path
        figure(22); clf; hold on;       
        %plot(y,x,'.','Color',0.8*[1 1 1],'Linewidth',2); % Nodes
        plot(y(idx_t),x(idx_t),'p','Color',[0.6350 0.0780 0.1840],...% Target 
            'MarkerFaceColor',[0.6350 0.0780 0.1840],'Markersize',8);
        plot(y(idx_o),x(idx_o),'ks','MarkerFaceColor','k'); % Obstacles
        [yGrid,xGrid] = meshgrid(0:dY:Ly,0:dX:Lx); % Grid
        plot(yGrid,xGrid,':','Color',0.8*[1 1 1]);
        plot(yGrid',xGrid',':','Color',0.8*[1 1 1]); 
        plot(optimal_path(:,2),optimal_path(:,1),'.-',... % Optimal path
            'Color',[0.4660 0.6740 0.1880],'LineWidth',1,'MarkerSize',10); 
        xlabel('y [m]'); ylabel('x [m]');
        axis equal tight, box on
        hold off;
    end

    
    %% Functions 
    function dist = distance(x1,y1,x2,y2)

        % This function calculates the distance between 
        % any two cartesian coordinates.
        %
        % Copyright 2009-2010 The MathWorks, Inc.

        dist = sqrt((x1-x2)^2 + (y1-y2)^2);

    end

    function exp_array = expand_array(psi, xPos, yPos, hn, xTarget,...
            yTarget, CLOSED, Lx, Ly, dX, dY, turnAngles, d, dPsi,...
            Z, heuristics)

        % Function generating an expanded array:
        % Taking a node as input, it returns the expanded list of 
        % successors, with the calculated total cost (fn). The criteria 
        % being none of the successors are on the CLOSED list.
        % 
        % Expanded Array Structure
        % ------------------------------------------
        % |xCell|yCell|psi|xPos|yPos|h(n)|g(n)|f(n)|
        % ------------------------------------------
        %
        % Copyright 2009-2010 The MathWorks, Inc.
        % Modified: Caroline Fleischer, 24.04.2020

        exp_array = nan(length(turnAngles),8);
        exp_count = 1;
        nClosed = size(CLOSED,1); % Number of elements in CLOSED
        for k = 1:length(turnAngles) % Generate child nodes
            beta = turnAngles(k);
            psiNew = ssa(psi+beta,'rad'); %[rad] New heading in (-pi,pi)
            xNew = xPos + d*cos(psiNew); %[m] New position
            yNew = yPos + d*sin(psiNew);
            if ((xNew >= 0 && xNew <= Lx) && (yNew >= 0 && yNew <= Ly))
                insert = true; % Node is within workspace
                [xNewCell,yNewCell] = findCell(xNew,yNew,dX,dY); % Finds which cell the new node is within 
                slope = (xNew-xPos)/(yNew-yPos); % Slope of the line segment
                if abs(slope)<1 % y is the major dimension
                    yMid = yNew-sign(yNew-yPos)*dY;%yPos+sign(yNew-yPos)*dY;
                    xMid = -slope*(yNew-yMid)+xNew;
                    [xMidCell,yMidCell] = findCell(xMid,yMid,dX,dY);
                else % x is the major dimension
                    xMid = xNew-sign(xNew-xPos)*dX;%xPos+sign(xNew-xPos)*dX;
                    yMid = -1/slope *(xNew-xMid)+yNew;
                    [xMidCell,yMidCell] = findCell(xMid,yMid,dX,dY);
                end 
                for c1 = 1:nClosed % Running through all closed nodes
                    if (xNewCell == CLOSED(c1,1) && yNewCell == CLOSED(c1,2)) % The cell is already closed
                        nClosedPsi = CLOSED(c1,3);
                        for p = 4:1:nClosedPsi+3 
                            if abs(psiNew - CLOSED(c1,p)) < 1e-5 % The cell is already closed for this heading (= there is already a path going through the cell with this heading which is cheaper)
                                insert = false;
                            end
                        end
                    elseif (xMidCell == CLOSED(c1,1) && yMidCell == CLOSED(c1,2) && CLOSED(c1,3) >= floor(2*pi/dPsi)+1) % The line segment passes through an obstructed cell
                        insert = false;
                    end % End of checking cell in between 
                end % End of checking whether child is in closed list
                if insert
                    % The cell where the node is located
                    exp_array(exp_count,1) = xNewCell;
                    exp_array(exp_count,2) = yNewCell;
                    % Path angle and node position
                    exp_array(exp_count,3) = psiNew;
                    exp_array(exp_count,4) = xNew;
                    exp_array(exp_count,5) = yNew;
                    % Cost of travelling to node, hn 
                    if heuristics
                        % Traveled distance adjusted for neural activity
                        exp_array(exp_count,6) = hn +...
                            0.5*(Z(yNewCell,xNewCell)+1)...
                            *distance(xPos,yPos,xNew,yNew);
                    else
                        % Traveled distance
                        exp_array(exp_count,6) = hn + distance(xPos,yPos,xNew,yNew);
                    end
                    % Cost from node to goal, gn (penalty on direction change and heading error)
                    if heuristics
                        % Heading error relative to progression straight towards the target
                        psi_e = abs(atan2(yTarget-yNew,xTarget-xNew)-psiNew);
                        % Target cell
                        [xTCell,yTCell] = findCell(xTarget,yTarget,dX,dY);
                        % Euclidean distance adjusted for neural activity
                        exp_array(exp_count,7) = abs(beta) + psi_e +...
                            0.5*(Z(yTCell,xTCell)+1)...
                            *distance(xTarget,yTarget,xNew,yNew);
                    else
                        % Euclidean distance
                        exp_array(exp_count,7) = distance(xTarget,yTarget,xNew,yNew);
                    end
                    % Total cost, fn
                    exp_array(exp_count,8) = exp_array(exp_count,6)...
                        + exp_array(exp_count,7); 
                    exp_count = exp_count + 1;
                end % Populate the exp_array list
            end % End of node within workspace     
        end % End of k for loop
        exp_array = exp_array(1:exp_count-1,:);
    
    end

    function [xCell,yCell] = findCell(x,y,dX,dY)

        % Function returning the corresponding cell coordinates to a position 
        % within the grid 
        %
        % Caroline Fleischer, 03.03.2020

        xCell = ceil(x/dX);
        yCell = ceil(y/dY);
    
    end

    function idx_min = min_fn(OPEN,OPEN_count,xTarget,yTarget,posAccuracy)

        % Function that returns the node with the smallest total 
        % cost, fn, or the goal node. It takes the OPEN List as input.
        %
        % Copyright 2009-2010 The MathWorks, Inc.

        temp_array = nan(OPEN_count,11);
        k = 1;
        % Iterate through all nodes in the open list
        for n = 1:OPEN_count
            if (OPEN(n,1) == 1)
                temp_array(k,:) = [OPEN(n,:) n];
                % Return the goal node if it is among the successors
                if norm([xTarget-OPEN(n,5), yTarget-OPEN(n,6)]) <= posAccuracy
                    idx_min = n; 
                    return;
                end 
                k = k + 1;
            end
        end 
        temp_array = temp_array(1:k-1,:);
        % Return the open list index of the node with the smallest f
        if size(temp_array ~= 0)
            [~,temp_min] = min(temp_array(:,10)); % Index in temp_array
            idx_min = temp_array(temp_min,11); % Index in the open list
        else
            idx_min = -1; % temp_array is empty (i.e. no feasible path)
        end

    end

    function roundedValue = roundValue(value,step)

        % Function that returns a value rounded to a certain resolution
        %
        % Adopted from Nordeus, E. (2019): Self-driving vehicle. Available 
        % at: https://github.com/Habrador/Self-driving-vehicle/blob/master...
        % /Self-driving%20vehicle%20Unity/Assets/Scripts/Pathfinding...
        % /Utility/HelpStuff.cs (accessed: 03.03.2020)

        roundedValue = round(value/step)*step;

    end

end