%Initialize and run simulation of CSE1 based on Fossen's kinematic 
%and kinetic vehicle model. The guidance system comprises a hybrid A* path 
%planner and a path generator based on Kndal's pragmatic method with 
%Bzier curves. 

clc, clear, close all
%clear 'hybridAstarBINN.m' 'hybridAstarBINN_kinematics.m' 
plot_prelims

%% User preferences for the guidance system
% Path planner
plot_search = 0; % plot hybrid A* search tree: 1 = on, 0 = off
plot_path   = 0; % plot hybrid A* path at each iteration: 1 = on, 0 = off

expansion   = 0; % 0 = straight-line, 1 = kinematics-based
heuristics  = 1; % 0 = Euclidean distance, 1 = adjusted for neural activity

% Path genertor
generator = 1; % 1 = pragmatic, 2 = optimal
if generator~=1 && generator ~=2
    error('The value of generator must be either 1 or 2.\n');
end

%% Angle conversion factors
rad2deg = 180/pi; % radians to degrees
deg2rad = pi/180; % degrees to radians

%% Simulation parameters
h = 0.01;     %[s] time step
Tstop = 1500; %[s] simulation time

%% CSE1 data
L   = 1.105;  % [m]  vessel length
m   = 14.11;  % [kg] vessel mass 
I_z = 1.7600; % [kg m2] moment of inertia 
x_g = 0.0375; % [m] x-coordinate of the center of gravity (y_g = 0)

X_ud = -2.0;  % added mass coefficients							
			 Y_vd = -10.0;	N_vd =  0.0;
			 Y_rd =   0.0;	N_rd = -1.0;

% Mass matrix
M    = [m - X_ud      0                0
          0         m - Y_vd     m*x_g - Y_rd
          0     m*x_g - N_vd       I_z - N_rd];

% Linear damping matrix
D_l  = -[-0.6555   0        0
          0       -1.330   -7.250
          0        0.0     -1.900];

%% Path-planning parameters
Uref  = 0.2;       %[m/s]   desired reference speed  %%NB: vessel is 1:50%%
Rmin  = 3/2 *L;    %[m]     minimum turn radius
r_max = Uref/Rmin; %[rad/s] maximum rotation rate

% Workspace parameters
Lx = 42;                    % [m] length in x-direction (north)
Ly = Lx;                    % [m] length in y-direction (east)

Nx = 14; dX = Lx/Nx;        % grid partitioning (regular)
Ny = Nx; dY = Ly/Ny;
N  = Nx*Ny;                 % total number of cells/neurons

[X,Y] = meshgrid(dX/2:dX:Lx,dY/2:dY:Ly);                 
x     = reshape(X,[N,1]);
y     = reshape(Y,[N,1]);
r0    = min(dX,dY)/6;       %[m] radius of acceptance (defining when the waypoint is regarded reached by the vessel)

% Neural network parameters
E    = 10;                  % constant for external input
d0   = max(dX,dY);          % [m] radius of the receptive field Ri (defining which are the neighboring neurons sqrt(dx^2+dy^2)
a    = 10;                  % passive decay rate (shunting eq.)
bu   = 1.0;                 % upper bound of neural activity (shunting eq.)            
bl   = 1.0;                 % lower bound of neural activity (shunting eq.)
cw   = 1; %10               % constant defining neural connection weights
e_ss = 1e-10;               % condition for steady-state neural activity 

% Matrices of neighbouring neurons A(i,j) and connection weights W(i,j)
[A,W] = neighborNeurons(x,y,N,d0,cw);

% Parameters for COLAV compliant with COLREGs
rd = 20;                    % [m] radius of the detection region
sd_L = 10*L;                % [m] large safe distance
sd_M = 5*L;                 % [m] medium safe distance
sd_S = 2*L;                 % [m] small safe distance

% Initial ASV position and heading
idx0 = Ny;                      % initial position index (lower right corner node)
psi0 = 0*deg2rad;               % [rad] initial heading (towards north)
eta0 = [x(idx0),y(idx0),psi0]'; % [m m rad]' initial state

% Target
TargetIdx = [Nx, 1];            % target index (upper left corner node)
[Nt,~] = size(TargetIdx);       % number of targets
idx_t = (TargetIdx(:,1)-1).*Ny ...
    + TargetIdx(:,2);           % indices of targets in vector grid          

% Static obstacles
StaticObIdx = [1,1; 2,1; 3,1; 1,2; 2,2; 6,6; 7,7; 8,8];
if ~isempty(StaticObIdx)
    idx_so = (StaticObIdx(:,1)-1).*Ny ...
        + StaticObIdx(:,2);     % indices of static obstacles in vector grid 
end

% Moving obstacle
r_obj   = 1;                              % [m] radius of object
p0_obj  = [33;31];                        % [m,m]' initial position                                   
U_obj   = 0.9*Uref;                       % [m/s] velocity
d0_obj  = pi;                             % [rad] initial direction 
theta   = linspace(0, 2*pi);
obj_def = r_obj*[cos(theta); sin(theta)]; % object definition (circle)

%% Observer parameters
b  = [0.3 0.1 0.02]'; % [N N Nm]' bias in surge, sway, and yaw zeros(3,1);%
vp = [20 20 0.04]'.*10^-5; % [m2s m2s rad2s]' power of white noise 

% Injection Gains
L1 = diag([1 1 1]);
L2 = diag([1 1 0.5]);%diag([0.1 0.1 0.02]);%1*eye(3);
L3 = 0.1*L2;%diag([0.1 0.1 0.05]);%0.1*eye(3); zeros(3);%

%% Path-generation parameters
n = 8; % number of control points (septic Bzier curve)

% Tuning variables
zeta = dX/2; %1 % [m] wall distance
delta_min = 2; %1; % [m] minimum distance from P4 to P7
my = 6; %6 % scaling factor for distance between P4, P5, P6, and P7
k_max = 3; %1/Rmin; %5; % [m-1] max curvature (optimal approach)
posib = 40; % size of the search space (optimal approach)
Delta_max = 2*dX; % [m] max distance between two WPs (optimal approach)

%% Motion control parameters
K1 = 0.01*eye(3); % control gain on z1 %Mar.Reg.II: 0.01 
K2 = 20*eye(3); % control gain on z2 %Mar.Reg.II: 10
mu = 0.1; %0.2 % scaling factor for speed assignment 
%(high mu = slow convergence; low mu = rapid convergence, but may overshoot) 

%% Run simulation
sim('CSE1_sim')

%% KPIs of the path generator
% length of the generated path
path_length = sum(sqrt(sum(diff(etad(:,1:2)).^2,2)));
% maximum curvature along the generated path
k_max_path = 0;
for j = 1:length(etad)
    k = norm(etad_s(j,1).*etad_s2(j,2)-etad_s2(j,1).*etad_s(j,2))/norm(etad_s(j,1:2))^3;
    k_max_path = max(k_max_path,k);
end
% average of execution time for generating desired setpoints
t_gen_avg = sum(t_gen(2:end))/(length(t_gen)-1);
% accumulated, average and max deviation from the path
diff_path = vecnorm(etad(2:end,1:2)-eta(2:end,1:2),2,2);
sum_diff_path = sum(diff_path);
avg_diff_path = sum_diff_path/(length(eta)-1);
max_diff_path = max(diff_path);

%% Plotting
figure(1); clf; hold on
plot([eta0(2);WP(:,2)],[eta0(1);WP(:,1)],'.','MarkerSize',12,'Color',green) % waypoints
plot(etad(:,2), etad(:,1),'-','Color',green) % desired Bezir path
plot(eta_obj(:,2), eta_obj(:,1),'-.','Color',orange) % moving object tracks
plot(eta(:,2), eta(:,1),'-k') % vessel tracks
plot(y(idx_so),x(idx_so),'ks','MarkerFaceColor','k'); % static obstacles
plot(y(idx_t),x(idx_t),'p','Color',red,'MarkerFaceColor',red); % target
[yGrid,xGrid] = meshgrid(0:dY:Ly,0:dX:Lx); % grid
plot(yGrid,xGrid,':','Color',gray);
plot(yGrid',xGrid',':','Color',gray);
axis([0 Ly 0 Lx]), axis equal tight manual, box on
for i = 1:Tstop/(50*h):length(eta)  % plot vessel
    R0 = [cos(eta(i,3)) -sin(eta(i,3)); 
          sin(eta(i,3))  cos(eta(i,3))];
    ship = R0*[6*L/9 5*L/9 3*L/9 -5*L/9 -5*L/9 3*L/9 5*L/9 6*L/9; 0 1*L/9 2*L/9 2*L/9 -2*L/9 -2*L/9 -1*L/9 0];
    plot(eta(i,2)+ship(2,:), eta(i,1)+ship(1,:),'k')
end
%Rd = rd * [cos(theta); sin(theta)] + eta(end,1:2)'; % detection region
%fill(Rd(2,:),Rd(1,:),'b','LineStyle','none','FaceAlpha',.05);
fill(obj_def(2,:)+eta_obj(end,2),obj_def(1,:)+eta_obj(end,1),... % moving object
    orange,'FaceAlpha',.2,'EdgeColor',orange);
legend('waypoints','B\''ezier path','object tracks','vessel tracks','static obstacles','target','Location','southeastoutside')
xlabel('y [m]'); xticks(dY:2*dY:Ly)
ylabel('x [m]'); yticks(dX:2*dX:Lx)
% Zoomed subplot: Create a new pair of axes inside current figure
axes('position',[.71 .49 .2 .2]) %[left bottom width height]
box on % put box around new pair of axes
idxOfInterest = (eta(:,2) > 32) & (eta(:,2) < 35); %(eta(:,2) > 28) & (eta(:,2) < 31);%
plot(eta(idxOfInterest,2),eta(idxOfInterest,1),'-k'); hold on 
plot(etad(idxOfInterest,2),etad(idxOfInterest,1),'-','Color',green)
%plot(WP(idxOfInterest,2),WP(idxOfInterest,1),'.','MarkerSize',12,'Color',green)
axis equal tight
hold off

figure(2); clf;
%title('Path parameter $s$')
plot(t, s, 'color', blue);
xlabel('$t$ [s]'); xlim([0 t(end)]);
ylabel('$s^\prime$ [-]');
grid on; box on;

figure(3); clf;
subplot(2,1,1); hold on;
%title('Desired, measured and estimated vessel position and heading')
plot(t, etad(:,1), 'color', darkgray, 'displayname', '$x_d$');%, 'linewidth', 2);
plot(t, eta_y(:,1), 'color', lblue, 'displayname', '{\boldmath${y}$}$_1$');
plot(t, eta_hat(:,1), 'color', red, 'displayname', '$\hat{x}$');
plot(t, etad(:,2), 'color', gray, 'displayname', '$y_d$');%, 'linewidth', 2);
plot(t, eta_y(:,2), 'color', green, 'displayname', '{\boldmath${y}$}$_2$');
plot(t, eta_hat(:,2), 'color', orange, 'displayname', '$\hat{y}$');
legend('show', 'numcolumns', 2, 'location', 'south');
xticklabels({}); xlim([0 t(end)]);
ylabel('\boldmath${p}$ [m]');
pos = get(gca, 'Position'); pos(2) = 0.55; pos(4) = 0.4; 
set(gca, 'Position', pos);
grid on; box on; hold off;
subplot(2,1,2); hold on;
plot(t, etad(:,3)*rad2deg, 'color', darkgray, 'displayname', '$\psi_d$');%, 'linewidth', 2);
plot(t, eta_y(:,3)*rad2deg, 'color', blue, 'displayname', '{\boldmath${y}$}$_3$');
plot(t, eta_hat(:,3)*rad2deg, 'color', orange, 'displayname', '$\hat{\psi}$');
legend('show','location','north')
xlabel('$t$ [s]'); xlim([0 t(end)]);
ylabel('$\psi$ [deg]');
pos = get(gca, 'Position'); pos(2) = 0.1; pos(4) = 0.4; 
set(gca, 'Position', pos);
grid on; box on; hold off;
 
figure(4); clf;
%title('Commanded forces and moment as function of time')
subplot(3,1,1); hold on;
plot(t, tau(:,1), 'color', red);
plot(t, 0.91*ones(length(t),1), 'color', orange, 'linewidth', 1, 'handlevisibility','off');
plot(t,-0.91*ones(length(t),1), 'color', orange, 'linewidth', 1, 'handlevisibility','off');
xticklabels({}); xlim([0 t(end)]);
ylabel('$X$ [N]');
pos = get(gca, 'Position'); pos(2) = 0.7; pos(4) = 0.25; 
set(gca, 'Position', pos);
grid on; box on; hold off;
subplot(3,1,2); hold on;
plot(t, tau(:,2), 'color', green);
plot(t, 0.5*ones(length(t),1), 'color', orange, 'linewidth', 1, 'handlevisibility','off');
plot(t,-0.5*ones(length(t),1), 'color', orange, 'linewidth', 1, 'handlevisibility','off');
xticklabels({}); xlim([0 t(end)]);
ylabel('$Y$ [N]'); ylim([-0.55 0.55]);
pos = get(gca, 'Position'); pos(2) = 0.4; pos(4) = 0.25; 
set(gca, 'Position', pos);
grid on; box on; hold off;
subplot(3,1,3); hold on;
plot(t, tau(:,3), 'color', blue); 
plot(t, 0.23*ones(length(t),1), 'color', orange, 'linewidth', 1, 'handlevisibility', 'off');
plot(t,-0.23*ones(length(t),1), 'color', orange, 'linewidth', 1, 'handlevisibility', 'off');
xlabel('$t$ [s]'); xlim([0 t(end)]);
ylabel('$N$ [Nm]'); ylim([-0.25 0.25]);
pos = get(gca, 'Position'); pos(2) = 0.1; pos(4) = 0.25; 
set(gca, 'Position', pos);
grid on; box on; hold off;

figure(5); clf;
%title('Estimated and real bias as function of time')
subplot(3,1,1); hold on;
plot(t, b(1).*ones(length(t),1), '--', 'color', red, 'displayname', '$b_X$');
plot(t, b_hat(:,1), 'color', red, 'displayname', '$\hat{b}_X$');
legend('show','Location','best');
xticklabels({}); xlim([0 t(end)]);
ylabel('$b_X$ [N]');
pos = get(gca, 'Position'); pos(2) = 0.7; pos(4) = 0.25; 
set(gca, 'Position', pos);
grid on; box on; hold off;
subplot(3,1,2); hold on;
plot(t, b(2).*ones(length(t),1), '--', 'color', green, 'displayname', '$b_Y$');
plot(t, b_hat(:,2), 'color', green, 'displayname', '$\hat{b}_Y$');
legend('show','Location','best');
xticklabels({}); xlim([0 t(end)]);
ylabel('$b_Y$ [N]');
pos = get(gca, 'Position'); pos(2) = 0.4; pos(4) = 0.25; 
set(gca, 'Position', pos);
grid on; box on; hold off;
subplot(3,1,3); hold on;
plot(t, b(3).*ones(length(t),1), '--', 'color', blue, 'displayname', '$b_N$');
plot(t, b_hat(:,3), 'color', blue, 'displayname', '$\hat{b}_N$');
legend('show','Location','best');
xlabel('$t$ [s]'); xlim([0 t(end)]);
ylabel('$b_N$ [Nm]');
pos = get(gca, 'Position'); pos(2) = 0.1; pos(4) = 0.25; 
set(gca, 'Position', pos);
grid on; box on; hold off;

figure(7); clf; 
subplot(3,1,1); hold on;
%title('Estimated and real velocity as function of time')
plot(t, nu(:,1), 'color', red, 'displayname', '$u$');
plot(t, nu_hat(:,1), '--', 'color', red, 'displayname', '$\hat{u}$');
legend('show','Location','best');
xticklabels({}); xlim([0 t(end)]);
ylabel('$u$ [m/s]');
pos = get(gca, 'Position'); pos(2) = 0.7; pos(4) = 0.25; 
set(gca, 'Position', pos);
grid on; box on; hold off;
subplot(3,1,2); hold on;
plot(t, nu(:,2), 'color', green, 'displayname', '$v$');
plot(t, nu_hat(:,2), '--', 'color', green, 'displayname', '$\hat{v}$');
legend('show','Location','best');
xticklabels({}); xlim([0 t(end)]);
ylabel('$v$ [m/s]');
pos = get(gca, 'Position'); pos(2) = 0.4; pos(4) = 0.25; 
set(gca, 'Position', pos);
grid on; box on; hold off;
subplot(3,1,3); hold on;
plot(t, nu(:,3)*rad2deg, 'color', blue, 'displayname', '$r$');
plot(t, nu_hat(:,3)*rad2deg, '--', 'color', blue, 'displayname', '$\hat{r}$');
legend('show','Location','best');
xlabel('$t$ [s]'); xlim([0 t(end)]);
ylabel('$r$ [deg/s]');
pos = get(gca, 'Position'); pos(2) = 0.1; pos(4) = 0.25; 
set(gca, 'Position', pos);
grid on; box on; hold off;

% figure(3); clf;
% plot(t, etad(:,3)*rad2deg, t, eta(:,3)*rad2deg, t, eta_hat(:,3)*rad2deg, ':')
% legend('$\psi_d$','$\psi$','$\psi_{est}$','Location','best')
% title('Vessel heading versus the desired and estimated heading')
% xlabel('$t$ [s]')
% ylabel('$\psi$ [deg]')
% xlim([0 t(end)])
% grid on

% figure(6); clf;
% plot(t, eta(:,1), 'r', t, eta_hat(:,1), 'r:',...
%     t, eta(:,2), 'g', t, eta_hat(:,2), 'g:')
% legend('$x$','$x_{est}$','$y$','$y_{est}$','Location','best')
% title('Estimated and real position as function of time')
% xlabel('$t$ [s]')
% ylabel('$p$ [m]')
% xlim([0 t(end)])
% grid on

% figure(7); clf; 
% plot(t, nu(:,1), 'r', t, nu(:,2), 'g', t, nu(:,3), 'b'); hold on
% plot(t, nu_hat(:,1), 'r--', t, nu_hat(:,2), 'g--', t, nu_hat(:,3), 'b--')
% legend('$u$','$v$','$r$','$u_{est}$','$v_{est}$','$r_{est}$','Location','best')
% title('Estimated and real velocity as function of time')
% xlabel('$t$ [s]')
% ylabel('$\nu$')
% xlim([0 t(end)])
% grid on
% hold off

% figure(4); clf;
% plot(t, tau(:,1), 'r', t, tau(:,2), 'g', t, tau(:,3), 'b'); hold on
% legend('$X$','$Y$','$N$','Location','best')
% plot(t, 0.91*ones(length(t),1), 'r--', t, -0.91*ones(length(t),1), 'r--',...
%     t, 0.5*ones(length(t),1), 'g--', t, -0.5*ones(length(t),1), 'g--',...
%     t, 0.23*ones(length(t),1), 'b--', t, -0.23*ones(length(t),1), 'b--',...
%     'HandleVisibility','off')
% title('Commanded forces and moment as function of time')
% xlabel('$t$ [s]')
% ylabel('$\tau$ [N/Nm]')
% xlim([0 t(end)])
% grid on
% hold off

%% Functions 
function [A,W] = neighborNeurons(x,y,N,d0,mu)
    % Determining neighbouring neurons A(i,j) in the BINN
    % and their connection weights W(i,j)
    % Input: 
    % x,y - coordinates of neurons in the network
    % N   - total number of neurons in the network
    % mu  - constant defining the connection between neurons
    A = zeros(N,N);     % Adjacency matrix
    W = zeros(N,N);     % Connection weights between neurons
    for i = 1:N
        for j = 1:N
            dist = max(abs(x(j)-x(i)),abs(y(j)-y(i)));
    %        dist = sqrt((x(j)-x(i))^2+(y(j)-y(i))^2);
            if (i ~= j) && (dist <= d0)
                A(i,j) = 1;
                W(i,j) = mu/dist;
            end
        end
    end
end
